package core.api.controller;

import core.api.models.auth.SimpleStatusResponse;
import core.api.models.timemanager.taskSchedule.*;
import core.entities.User;
import core.entities.timemanager.BasicTaskSchedule;
import core.entities.timemanager.ScheduleType;
import core.entities.timemanager.Task;
import core.repositories.UserRepository;
import core.repositories.timemanager.BasicTaskScheduleRepository;
import core.services.*;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.util.*;

@CrossOrigin(origins = "*", maxAge = 3600)
@RestController
@RequestMapping("/api")
public class ScheduleController {

    private final TaskService taskService;
    private final TaskScheduleService taskScheduleService;
    private final BasicTaskScheduleRepository basicTaskScheduleRepository;
    private final UserRepository userRepository;

    public ScheduleController(@Autowired TaskService taskService, @Autowired TaskScheduleService taskScheduleService,
                              BasicTaskScheduleRepository basicTaskScheduleRepository,
                              UserRepository userRepository) {
        this.taskService = taskService;
        this.taskScheduleService = taskScheduleService;
        this.basicTaskScheduleRepository = basicTaskScheduleRepository;
        this.userRepository = userRepository;
    }

    @GetMapping("/schedules/{taskID}/{scheduleType}")
    public ResponseEntity<?> loadSchedulesOfTask(@PathVariable long taskID, @PathVariable ScheduleType scheduleType) {
        PermissionResult<Task> taskPermissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!taskPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(taskPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        switch (scheduleType) {
            case BASIC -> {
                List<BasicTaskSchedule> basicTaskScheduleList = new ArrayList<>(taskPermissionResult.getResult().getBasicTaskSchedules());
                basicTaskScheduleList.sort(Comparator.comparing(BasicTaskSchedule::getScheduleDate));
                return ResponseEntity.ok(basicTaskScheduleList.stream().map(BasicTaskScheduleEntityInfo::new));
            }
            case ADVANCED, MODERATE -> ResponseEntity.ok(new ArrayList<>());
        }
        return null;
    }

    @PutMapping("/schedules/{taskID}/basic")
    public ResponseEntity<?> basicScheduleTask(@PathVariable long taskID, @Valid @RequestBody BasicTaskScheduleFieldInfo basicTaskSchedule) {
        PermissionResult<Task> taskPermissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!taskPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(taskPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<BasicTaskSchedule> creationResult = taskScheduleService.createBasicTaskSchedule(taskPermissionResult.getResult(), basicTaskSchedule);
        return ResponseEntity.ok(new BasicTaskScheduleEntityInfo(creationResult.getResult()));
    }

    @PostMapping("/schedules/{scheduleID}/basic")
    public ResponseEntity<?> editScheduledTask(@PathVariable long scheduleID, @Valid @RequestBody BasicTaskScheduleFieldInfo basicTaskScheduleFieldInfo) {
        PermissionResult<BasicTaskSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        taskScheduleService.reschedule(permissionResult.getResult(), basicTaskScheduleFieldInfo);
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }

    @DeleteMapping("/schedules/{scheduleID}/{scheduleType}")
    public ResponseEntity<?> deleteScheduledTask(@PathVariable long scheduleID, @PathVariable ScheduleType scheduleType) {
        PermissionResult<BasicTaskSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        taskScheduleService.deleteBasicSchedule(permissionResult.getResult());
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }

    @GetMapping("/schedules/today/{activateable}")
    public ResponseEntity<?> loadTodaysSchedules(@PathVariable boolean activateable) {
        ServiceResult<List<BasicTaskSchedule>> todaysSchedules = taskScheduleService.loadTodaysSchedule(SecurityContextHolder.getContext().getAuthentication().getName(), activateable);
        if(todaysSchedules.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        } else {
            return ResponseEntity.ok(todaysSchedules.getResult().stream().map(ScheduleInfo::new).toList());
        }
    }

    @GetMapping("/schedules")
    public ResponseEntity<?> loadSchedules() {
        ServiceResult<List<BasicTaskSchedule>> schedules = taskScheduleService.loadSchedules(SecurityContextHolder.getContext().getAuthentication().getName());
        if(schedules.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        } else {
            return ResponseEntity.ok(schedules.getResult().stream().map(ScheduleInfo::new).toList());
        }
    }
    @PostMapping("/schedules/{taskID}/now")
    public ResponseEntity<?> scheduleTaskNow(@PathVariable long taskID) {
        PermissionResult<Task> permissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<BasicTaskSchedule> scheduleResult = taskScheduleService.scheduleTaskNow(permissionResult.getResult());
        if(scheduleResult.getExitCode() == ServiceExitCode.ENTITY_ALREADY_EXIST) {
            return ResponseEntity.status(409).body(new SimpleStatusResponse("failed"));
        } else {
            return ResponseEntity.ok(new ScheduleInfo(scheduleResult.getResult()));
        }
    }

    @GetMapping("/schedules/active")
    public ResponseEntity<?> getActiveSchedule() {
        ServiceResult<BasicTaskSchedule> activeScheduleResult = taskScheduleService.getActiveSchedule(SecurityContextHolder.getContext().getAuthentication().getName());
        if(activeScheduleResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.ok(new ScheduleInfo());
        } else {
            return ResponseEntity.ok(new ScheduleInfo(activeScheduleResult.getResult()));
        }
    }

    @PostMapping("/schedules/{scheduleID}/activate")
    public ResponseEntity<?> activateSchedule(@PathVariable long scheduleID) {
        PermissionResult<BasicTaskSchedule> schedulePermissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!schedulePermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(schedulePermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<BasicTaskSchedule> scheduleActivateResult = taskScheduleService.activateSchedule(schedulePermissionResult.getResult());
        if(scheduleActivateResult.getExitCode() == ServiceExitCode.ENTITY_ALREADY_EXIST) {
            return ResponseEntity.status(409).body(new SimpleStatusResponse("failed"));
        } else if(scheduleActivateResult.getExitCode() == ServiceExitCode.OK){
            return ResponseEntity.ok(new ScheduleActivateInfo(scheduleActivateResult.getResult().getStartTime()));
        } else {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }
    }

    @PostMapping("/schedules/{scheduleID}/stop/{finish}")
    public ResponseEntity<?> stopSchedule(@PathVariable long scheduleID, @PathVariable boolean finish) {
        PermissionResult<BasicTaskSchedule> schedulePermissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!schedulePermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(schedulePermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<Integer> stopResult = taskScheduleService.stopSchedule(schedulePermissionResult.getResult(), finish);
        if(stopResult.getExitCode() == ServiceExitCode.INVALID_OPERATION) {
            return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
        } else {
            return ResponseEntity.ok(new TaskScheduleStopResponse(stopResult.getResult()));
        }
    }

    @GetMapping("/schedules/status/today")
    public ResponseEntity<?> getWorkedMinutesToday() {
        Optional<User> user = userRepository.findByUsername(SecurityContextHolder.getContext().getAuthentication().getName());
        if(user.isEmpty()) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        int workedMinutes = taskScheduleService.getWorkedMinutes(user.get());
        return ResponseEntity.ok(new ScheduleStatus(workedMinutes, taskScheduleService.isAnyScheduleMissed(user.get())));
    }

    @PostMapping("/schedules/{taskID}/forgotten")
    public ResponseEntity<?> registerForgottenActivity(@PathVariable long taskID, @Valid @RequestBody ForgottenActivity forgottenActivity) {
        PermissionResult<Task> permissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<Integer> serviceResult = taskScheduleService.registerForgottenActivity(permissionResult.getResult(), forgottenActivity);
        if(serviceResult.getExitCode() == ServiceExitCode.INVALID_OPERATION) {
            return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
        } else {
            return ResponseEntity.ok(new TaskScheduleStopResponse(serviceResult.getResult()));
        }
    }

    @GetMapping("/schedules/missed")
    public ResponseEntity<?> loadMissedSchedules() {
        Optional<User> user = userRepository.findByUsername(SecurityContextHolder.getContext().getAuthentication().getName());
        if(user.isEmpty()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        List<BasicTaskSchedule> schedules = taskScheduleService.loadMissedSchedules(user.get());
        return ResponseEntity.ok(schedules.stream().map(ScheduleInfo::new).toList());
    }

    @DeleteMapping("/schedules/{scheduleIDs}/all")
    public ResponseEntity<?> deleteSchedules(@PathVariable long[] scheduleIDs) {
        List<BasicTaskSchedule> schedulesToDelete = new LinkedList<>();
        for(long scheduleID: scheduleIDs) {
            PermissionResult<BasicTaskSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
            if(!permissionResult.isHasPermissions()) {
                return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
            }

            if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
               return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
            }
            schedulesToDelete.add(permissionResult.getResult());
        }

        for(BasicTaskSchedule schedule : schedulesToDelete) {
            this.taskScheduleService.deleteBasicSchedule(schedule);
        }
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }
}
