package core.services;

import core.api.models.timemanager.taskSchedule.BasicTaskScheduleFieldInfo;
import core.api.models.timemanager.taskSchedule.ForgottenActivity;
import core.entities.User;
import core.entities.timemanager.BasicTaskSchedule;
import core.entities.timemanager.Task;
import core.repositories.UserRepository;
import core.repositories.timemanager.BasicTaskScheduleRepository;
import core.repositories.timemanager.TaskRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;

@Service
public class TaskScheduleService {

    private final BasicTaskScheduleRepository basicTaskScheduleRepository;
    private final TaskRepository taskRepository;
    private final UserRepository userRepository;

    public TaskScheduleService(@Autowired BasicTaskScheduleRepository basicTaskScheduleRepository,
                               @Autowired TaskRepository taskRepository,
                               UserRepository userRepository) {
        this.basicTaskScheduleRepository = basicTaskScheduleRepository;
        this.taskRepository = taskRepository;
        this.userRepository = userRepository;
    }

    public ServiceResult<BasicTaskSchedule> createBasicTaskSchedule(Task task, BasicTaskScheduleFieldInfo basicTaskScheduleFieldInfo) {
        BasicTaskSchedule basicTaskSchedule = new BasicTaskSchedule(task, basicTaskScheduleFieldInfo.getScheduleDate());
        task.getBasicTaskSchedules().add(basicTaskSchedule);
        basicTaskScheduleRepository.save(basicTaskSchedule);
        taskRepository.save(task);
        return new ServiceResult<>(basicTaskSchedule);
    }

    public PermissionResult<BasicTaskSchedule> getSchedulePermissions(long scheduleID, String name) {
        Optional<BasicTaskSchedule> basicSchedule = basicTaskScheduleRepository.findById(scheduleID);
        return basicSchedule.map(value -> new PermissionResult<>(value, value.getTask().getTaskgroup().getUser().getUsername().equals(name))).orElseGet(() ->
                new PermissionResult<>(ServiceExitCode.MISSING_ENTITY));
    }

    public void reschedule(BasicTaskSchedule basicTaskSchedule, BasicTaskScheduleFieldInfo basicTaskScheduleFieldInfo) {
        basicTaskSchedule.setScheduleDate(basicTaskScheduleFieldInfo.getScheduleDate());
        basicTaskScheduleRepository.save(basicTaskSchedule);
    }

    public void deleteBasicSchedule(BasicTaskSchedule basicTaskSchedule) {
        basicTaskScheduleRepository.deleteBasicTaskScheduleByID(basicTaskSchedule.getScheduleID());
    }

    public ServiceResult<List<BasicTaskSchedule>> loadTodaysSchedule(String username, boolean onlyActivateable) {
        Optional<User> user = userRepository.findByUsername(username);
        if(user.isEmpty()) {
            return new ServiceResult<>(ServiceExitCode.MISSING_ENTITY);
        }

        List<BasicTaskSchedule> basicTaskSchedules = basicTaskScheduleRepository.findAllByUserAndDate(user.get(), LocalDate.now());
        List<BasicTaskSchedule> activatableSchedules = new LinkedList<>();
        if(onlyActivateable) {
            for (BasicTaskSchedule basicTaskSchedule : basicTaskSchedules) {
                if (basicTaskSchedule.isActivateAble()) {
                    activatableSchedules.add(basicTaskSchedule);
                }
            }
        }
        return new ServiceResult<>(activatableSchedules);
    }

    public ServiceResult<List<BasicTaskSchedule>> loadSchedules(String username) {
        Optional<User> user = userRepository.findByUsername(username);
        return user.map(value -> new ServiceResult<>(basicTaskScheduleRepository.findAllByUser(value))).orElseGet(() ->
                new ServiceResult<>(ServiceExitCode.MISSING_ENTITY));
    }

    public ServiceResult<BasicTaskSchedule> scheduleTaskNow(Task task) {
        //Check if task has already an active schedule
        ServiceResult<BasicTaskSchedule> activeSchedule = getActiveSchedule(task.getTaskgroup().getUser().getUsername());
        if(task.hasActiveSchedule() || activeSchedule.getExitCode() != ServiceExitCode.MISSING_ENTITY) {
            return new ServiceResult<>(ServiceExitCode.ENTITY_ALREADY_EXIST);
        } else {
            BasicTaskSchedule basicTaskSchedule = new BasicTaskSchedule(task, LocalDate.now());
            basicTaskSchedule.setStartTime(LocalDateTime.now());
            task.getBasicTaskSchedules().add(basicTaskSchedule);
            basicTaskScheduleRepository.save(basicTaskSchedule);
            taskRepository.save(task);
            return new ServiceResult<>(basicTaskSchedule);
        }
    }

    public void deleteScheduleByTask(Task task) {
        basicTaskScheduleRepository.deleteAllByTask(task);
    }

    public ServiceResult<BasicTaskSchedule> getActiveSchedule(String username) {
        Optional<BasicTaskSchedule> activeTaskScheduleResult = basicTaskScheduleRepository.findActiveTaskSchedule(username);
        return activeTaskScheduleResult.map(ServiceResult::new).orElseGet(() -> new ServiceResult<>(ServiceExitCode.MISSING_ENTITY));
    }

    public ServiceResult<BasicTaskSchedule> activateSchedule(BasicTaskSchedule taskSchedule) {
        //Check if taskSchedule can be started
        if(taskSchedule.getStartTime() != null) {
            return new ServiceResult<>(ServiceExitCode.INVALID_OPERATION);
        }

        if(getActiveSchedule(taskSchedule.getTask().getTaskgroup().getUser().getUsername()).getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            taskSchedule.setStartTime(LocalDateTime.now());
            basicTaskScheduleRepository.save(taskSchedule);
            return new ServiceResult<>(taskSchedule);
        } else {
            return new ServiceResult<>(ServiceExitCode.ENTITY_ALREADY_EXIST);
        }
    }

    public ServiceResult<Integer> stopSchedule(BasicTaskSchedule taskSchedule, boolean finish) {
        if(taskSchedule.getStartTime() == null || taskSchedule.getFinishedTime() != null) {
            return new ServiceResult<>(ServiceExitCode.INVALID_OPERATION);
        }

        taskSchedule.setFinishedTime(LocalDateTime.now());
        long activeTime = Duration.between(taskSchedule.getStartTime(), taskSchedule.getFinishedTime()).toMinutes();
        long workTime = taskSchedule.getTask().getWorkTime() + activeTime;
        taskSchedule.getTask().setWorkTime((int) workTime);
        if(finish) {
            taskSchedule.getTask().setFinished(true);
        }
        basicTaskScheduleRepository.save(taskSchedule);
        taskRepository.save(taskSchedule.getTask());
        return new ServiceResult<>((int) activeTime);
    }

    public int getWorkedMinutes(User user) {
        List<BasicTaskSchedule> basicTaskSchedules = basicTaskScheduleRepository.findAllByUser(user);
        long workedMinutes = 0;
        for(BasicTaskSchedule basicTaskSchedule : basicTaskSchedules) {
            if(basicTaskSchedule.getFinishedTime() != null && basicTaskSchedule.getFinishedTime().toLocalDate().isEqual(LocalDate.now())) {
                workedMinutes += Duration.between(basicTaskSchedule.getStartTime(), basicTaskSchedule.getFinishedTime()).toMinutes();
            }
        }
        return (int) workedMinutes;
    }

    public boolean isAnyScheduleMissed(User user) {
        List<BasicTaskSchedule> unstartedSchedules = basicTaskScheduleRepository.findAllByStartTimeIsNull();
        LocalDate dateReference = LocalDate.now();
        for(BasicTaskSchedule schedule : unstartedSchedules) {
            if(schedule.getScheduleDate().isBefore(dateReference)) {
                return true;
            }
        }
        return false;
    }

    public ServiceResult<Integer> registerForgottenActivity(Task task, ForgottenActivity forgottenActivity) {
        switch (forgottenActivity.getMode()) {
            case MANUAL -> {
                task.increaseActiveTime(forgottenActivity.getMinutesSpent());
                BasicTaskSchedule basicTaskSchedule = new BasicTaskSchedule(task, LocalDate.now());
                LocalDateTime timeReference = LocalDateTime.now();
                basicTaskSchedule.setStartTime(timeReference.minusMinutes(forgottenActivity.getMinutesSpent()));
                basicTaskSchedule.setFinishedTime(timeReference);
                task.getBasicTaskSchedules().add(basicTaskSchedule);
                basicTaskScheduleRepository.save(basicTaskSchedule);
                taskRepository.save(task);
                return new ServiceResult<>(forgottenActivity.getMinutesSpent());
            }
            case LAST -> {
                List<BasicTaskSchedule> schedules = basicTaskScheduleRepository.findAllFinishedByUserAndDate(task.getTaskgroup().getUser(), LocalDate.now());
                if(schedules.isEmpty()) {
                    return new ServiceResult<>(ServiceExitCode.INVALID_OPERATION);
                } else {
                    LocalDateTime timeReference = LocalDateTime.now();
                    BasicTaskSchedule nearestSchedule = null;
                    long nearestDuration = Long.MAX_VALUE;
                    for(BasicTaskSchedule schedule : schedules) {
                        if(nearestSchedule == null) {
                            nearestSchedule = schedule;
                            nearestDuration = Duration.between(nearestSchedule.getFinishedTime(), timeReference).getSeconds();
                        } else {
                            long currentDuration = Duration.between(schedule.getFinishedTime(), timeReference).getSeconds();
                            if(currentDuration < nearestDuration) {
                                nearestSchedule = schedule;
                                nearestDuration = currentDuration;
                            }
                        }
                    }

                    int minutesSpent = (int) Duration.between(nearestSchedule.getFinishedTime(), timeReference).toMinutes();
                    task.increaseActiveTime(minutesSpent);
                    BasicTaskSchedule basicTaskSchedule = new BasicTaskSchedule(task, timeReference.toLocalDate());
                    basicTaskSchedule.setStartTime(nearestSchedule.getFinishedTime());
                    basicTaskSchedule.setFinishedTime(timeReference);
                    basicTaskScheduleRepository.save(basicTaskSchedule);
                    taskRepository.save(task);
                    return new ServiceResult<>(minutesSpent);
                }
            }
            case PLANNED -> {
                //Does not make sense until advanced schedule/moderate schedule is implemented
                return new ServiceResult<>(0);
            }
            default -> {
                throw new RuntimeException("INVALID MODE");
            }
        }

    }

    public List<BasicTaskSchedule> loadMissedSchedules(User user) {
        List<BasicTaskSchedule> unstartedSchedules = basicTaskScheduleRepository.findAllUnstartedSchedulesOfUser(user);
        LocalDate currentDate = LocalDate.now();
        List<BasicTaskSchedule> missedSchedules = new LinkedList<>();
        for(BasicTaskSchedule schedule : unstartedSchedules) {
            if(schedule.getScheduleDate().isBefore(currentDate)) {
                missedSchedules.add(schedule);
            }
        }
        return missedSchedules;
    }
}
