package core.tasks;

import core.api.models.timemanager.tasks.TaskFieldInfo;
import core.entities.timemanager.Task;
import core.entities.timemanager.Taskgroup;
import core.repositories.timemanager.TaskRepository;
import core.repositories.timemanager.TaskgroupRepository;
import core.services.*;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.jdbc.Sql;
import org.springframework.test.context.jdbc.SqlGroup;

import javax.persistence.EntityManager;
import javax.transaction.Transactional;

import java.time.LocalDate;

import static org.assertj.core.api.AssertionsForClassTypes.assertThat;
import static org.junit.jupiter.api.Assertions.*;

@SpringBootTest
@Transactional
public class TaskServiceTest {


    @Autowired
    private TaskService taskService;
    @Autowired
    private EntityManager entityManager;

    private static final String username = "Testuser1";
    private static final String username2 = "Testuser2";
    @Autowired
    private TaskRepository taskRepository;
    @Autowired
    private TaskgroupRepository taskgroupRepository;

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql")
    })
    @Rollback(value = true)
    void createTask() {
        //Situation 1: Task with that name already exists in a taskgroup
        TaskFieldInfo taskFieldInfo_1 = new TaskFieldInfo(entityManager.find(Task.class, 1L).getTaskName(), 0, null, null);
        ServiceResult<Task> situation_1 = taskService.createTask(entityManager.find(Taskgroup.class, 2L), taskFieldInfo_1);
        assertEquals(ServiceExitCode.ENTITY_ALREADY_EXIST, situation_1.getExitCode());

        //Situation 2: Task with that name exists in another taskgroup
        TaskFieldInfo taskFieldInfo_2 = new TaskFieldInfo(entityManager.find(Task.class, 1L).getTaskName(), 0, null, null);
        ServiceResult<Task> result_2 = taskService.createTask(entityManager.find(Taskgroup.class, 1L), taskFieldInfo_2);
        assertEquals(ServiceExitCode.OK, result_2.getExitCode());
        assertThat(entityManager.find(Task.class, result_2.getResult().getTaskID())).isNotNull();

        //Situation 3: Normal, everything fine
        TaskFieldInfo taskFieldInfo_3 = new TaskFieldInfo("Situation 3", 0, null, null);
        ServiceResult<Task> result_3 = taskService.createTask(entityManager.find(Taskgroup.class, 1L), taskFieldInfo_3);
        assertEquals(ServiceExitCode.OK, result_3.getExitCode());
        assertThat(entityManager.find(Task.class, result_3.getResult().getTaskID())).isNotNull();

        //Robustness - Test for invalid dates
        //Situation 4: Deadline, no start
        TaskFieldInfo taskFieldInfo_4 = new TaskFieldInfo("Situation 4", 0, null, LocalDate.now());
        ServiceResult<Task> result_4 = taskService.createTask(entityManager.find(Taskgroup.class, 1L), taskFieldInfo_4);
        assertEquals(ServiceExitCode.OK, result_4.getExitCode());
        assertThat(entityManager.find(Task.class, result_4.getResult().getTaskID())).isNotNull();

        //Situation 5: Start but no deadline
        TaskFieldInfo taskFieldInfo_5 = new TaskFieldInfo("Situation 5", 0, LocalDate.now(),null);
        ServiceResult<Task> result_5 = taskService.createTask(entityManager.find(Taskgroup.class, 1L), taskFieldInfo_5);
        assertEquals(ServiceExitCode.OK, result_5.getExitCode());
        assertThat(entityManager.find(Task.class, result_5.getResult().getTaskID())).isNotNull();

        //Situation 6: Deadline before start (invalid)
        TaskFieldInfo taskFieldInfo_6 = new TaskFieldInfo("Situation 6", 0, LocalDate.now(),LocalDate.of(2010, 3, 20));
        ServiceResult<Task> result_6= taskService.createTask(entityManager.find(Taskgroup.class, 1L), taskFieldInfo_6);
        assertEquals(ServiceExitCode.INVALID_PARAMETER, result_6.getExitCode());
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql")
    })
    void getTaskPermissions() {
        //Situation 1: correct task and username
        PermissionResult<Task> permissionResult_1 = taskService.getTaskPermissions(1L, username);
        assertEquals(ServiceExitCode.OK, permissionResult_1.getExitCode());
        assertTrue(permissionResult_1.isHasPermissions());
        assertEquals(entityManager.find(Task.class, 1L), permissionResult_1.getResult());

        //Situation 2: invalid taskgroup
        PermissionResult<Task> permissionResult_2 = taskService.getTaskPermissions(200L, username);
        assertEquals(ServiceExitCode.MISSING_ENTITY, permissionResult_2.getExitCode());
        assertFalse(permissionResult_2.isHasPermissions());
        assertThat(permissionResult_2.getResult()).isNull();

        //Situation 3: invalid user
        PermissionResult<Task> permissionResult_3 = taskService.getTaskPermissions(1L, username2);
        assertEquals(ServiceExitCode.OK, permissionResult_3.getExitCode());
        assertFalse(permissionResult_2.isHasPermissions());
        assertEquals(entityManager.find(Task.class, 1L), permissionResult_3.getResult());
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql")
    })
    void editTask() {
        //Situation 1: Nothing is updated
        Task task_1 = entityManager.find(Task.class, 1L);
        TaskFieldInfo taskFieldInfo_1 = new TaskFieldInfo(entityManager.find(Task.class, 1L).getTaskName(), 0, null, null);
        ServiceResult<Task> result_1 = taskService.editTask(entityManager.find(Task.class, 1L), taskFieldInfo_1);
        assertEquals(ServiceExitCode.OK, result_1.getExitCode());
        assertEquals(task_1, result_1.getResult());

        //Situation 2: Name is updated to name that already exists within another taskgroup
        Task task_2 = entityManager.find(Task.class, 1L);
        task_2.setTaskName(entityManager.find(Task.class, 2L).getTaskName());
        TaskFieldInfo taskFieldInfo_2 = new TaskFieldInfo(entityManager.find(Task.class, 2L).getTaskName(), 0, null, null);
        ServiceResult<Task> result_2 = taskService.editTask(entityManager.find(Task.class, 1L), taskFieldInfo_2);
        assertEquals(ServiceExitCode.ENTITY_ALREADY_EXIST, result_2.getExitCode());
        assertNotEquals(task_2, result_2.getResult());

        //Situation 3: Name is updated to nonexistend Name
        Task task_3 = entityManager.find(Task.class, 1L);
        task_3.setTaskName("Situation 3");
        TaskFieldInfo taskFieldInfo_3 = new TaskFieldInfo("Situation 3", 0, null, null);
        ServiceResult<Task> result_3 = taskService.editTask(entityManager.find(Task.class, 1L), taskFieldInfo_3);
        assertEquals(ServiceExitCode.OK, result_3.getExitCode());
        assertEquals(task_3, result_3.getResult());

        //Situation 4: eta is updated
        Task task_4 = entityManager.find(Task.class, 1L);
        task_4.setEta(90);
        TaskFieldInfo taskFieldInfo_4 = new TaskFieldInfo(task_4.getTaskName(), 90, null, null);
        ServiceResult<Task> result_4 = taskService.editTask(entityManager.find(Task.class, 1L), taskFieldInfo_4);
        assertEquals(ServiceExitCode.OK, result_4.getExitCode());
        assertEquals(task_4, result_4.getResult());

        //Situation 5: deadline and start is updated in a valid way
        Task task_5 = entityManager.find(Task.class, 1L);
        task_5.setStartDate(LocalDate.of(2023,4,9));
        task_5.setDeadline(LocalDate.of(2023,5,9));
        TaskFieldInfo taskFieldInfo_5 = new TaskFieldInfo(task_5.getTaskName(), task_5.getEta(), LocalDate.of(2023, 4, 9), LocalDate.of(2023,5,9));
        ServiceResult<Task> result_5 = taskService.editTask(entityManager.find(Task.class, 1L), taskFieldInfo_5);
        assertEquals(ServiceExitCode.OK, result_5.getExitCode());
        assertEquals(task_5, result_5.getResult());

        //Situation 6: Deadline and start are updated in a not valid way
        Task task_6 = entityManager.find(Task.class, 1L);
        task_6.setStartDate(LocalDate.of(2023,5,9));
        task_6.setDeadline(LocalDate.of(2023,4,9));
        TaskFieldInfo taskFieldInfo_6 = new TaskFieldInfo(task_5.getTaskName(), task_5.getEta(), LocalDate.of(2023, 5, 9), LocalDate.of(2023,4,9));
        ServiceResult<Task> result_6 = taskService.editTask(entityManager.find(Task.class, 1L), taskFieldInfo_6);
        assertEquals(ServiceExitCode.INVALID_PARAMETER, result_6.getExitCode());
        assertNotEquals(task_6, result_6.getResult());

        //Situation 10 Startdate = null;
        Task task_10 = entityManager.find(Task.class, 5L);
        task_10.setStartDate(null);
        TaskFieldInfo taskFieldInfo_10 = new TaskFieldInfo(task_10.getTaskName(), task_10.getEta(),null, task_10.getDeadline());
        ServiceResult<Task> result_10 = taskService.editTask(entityManager.find(Task.class, 5L), taskFieldInfo_10);
        assertEquals(ServiceExitCode.OK, result_10.getExitCode());
        assertEquals(task_10, result_10.getResult());

        //Situation 11 Deadline = null
        Task task_11 = entityManager.find(Task.class, 5L);
        task_11.setDeadline(null);
        TaskFieldInfo taskFieldInfo_11 = new TaskFieldInfo(task_11.getTaskName(), task_11.getEta(),task_11.getStartDate(), null);
        ServiceResult<Task> result_11 = taskService.editTask(entityManager.find(Task.class, 5L), taskFieldInfo_11);
        assertEquals(ServiceExitCode.OK, result_11.getExitCode());
        assertEquals(task_11, result_11.getResult());
    }
    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql")
    })
    void deleteTask() {
        Task task = entityManager.find(Task.class, 1L);
        task.getTaskgroup().getTasks().remove(task);
        taskgroupRepository.save(task.getTaskgroup());
        Task updatedTask = entityManager.find(Task.class, 1L);
        taskRepository.deleteByTaskID(entityManager.find(Task.class, 1L).getTaskID());
        assertThat(entityManager.find(Task.class, 1L)).isNull();
    }
}
